/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /home/mobiserve2/mjmille2/CVS_REPOSITORY/ns-2.26/mac/Attic/mac-802_11m.cc,v 1.1.2.15 2003/05/01 22:49:15 mjmille2 Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 */

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

 #define DEBUG 99

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-m-timers.h"
#include "mac-802_11m.h"
#include "cmu-trace.h"
// MATT
// class MacLow;
#include "mac-low.h"
// end MATT


// XXX Can't we make these macros inline methods? Otherwise why should we have
// inline methods at all??

#define CHECK_BACKOFF_TIMER()						\
{									\
	if(is_idle() && mhBackoff_.paused())				\
		mhBackoff_.resume(difs_);				\
	if(! is_idle() && mhBackoff_.busy() && ! mhBackoff_.paused())	\
		mhBackoff_.pause();					\
}

#define TRANSMIT(p, t)                                                  \
{                                                                       \
	tx_active_ = 1;                                                  \
                                                                        \
        /*                                                              \
         * If I'm transmitting without doing CS, such as when           \
         * sending an ACK, any incoming packet will be "missed"         \
         * and hence, must be discarded.                                \
         */                                                             \
        if(rx_state_ != MAC_IDLE) {                                      \
                struct hdr_mac802_11m *dh = HDR_MAC802_11M(p);                  \
                                                                        \
                assert(dh->dh_fc.fc_type == MAC_Type_Control);          \
                assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);        \
                                                                        \
                assert(pktRx_);                                          \
                struct hdr_cmn *ch = HDR_CMN(pktRx_);                    \
                                                                        \
                ch->error() = 1;        /* force packet discard */      \
        }                                                               \
                                                                        \
        /*                                                              \
         * pass the packet on the "interface" which will in turn        \
         * place the packet on the channel.                             \
         *                                                              \
         * NOTE: a handler is passed along so that the Network          \
         *       Interface can distinguish between incoming and         \
         *       outgoing packets.                                      \
         */                                                             \
        downtarget_->recv(p->copy(), this);                             \
                                                                        \
			fprintf(stderr, "%.9f %d starting send timer\n",NOW,index_); \
        mhSend_.start(t);                                                \
                                                                        \
			fprintf(stderr, "%.9f %d starting IF timer\n",NOW,index_); \
	mhIF_.start(txtime(p));                                         \
}

#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);		\
	CHECK_BACKOFF_TIMER();		\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);			\
	CHECK_BACKOFF_TIMER();			\
}

/* ======================================================================
   Global Variables
   ====================================================================== */

static PHY_MIB PMIB =
{
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength, DSSS_PLCPDataRate
};	

static MAC_MIB MMIB =
{
	MAC_RTSThreshold, MAC_ShortRetryLimit,
	MAC_LongRetryLimit, MAC_FragmentationThreshold,
	MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

// MATT

void MacTransTimer::expire(Event *) {
	m_->sendPktWait();
}

void SleepTimer::expire(Event *) {
	m_->sleepHandler();
}

void NextWakeTimer::expire(Event *) {
	m_->nextWakeHandler();
}

void Mac802_11M::sendPktWait() {
	if(pktWait_ != 0)
		send(pktWait_, callback_); 
}

void
Mac802_11M::sleepHandler()
{
	if(netif_->is_on())
		netif_->node_sleep();

	fprintf(stderr, "%.9f %d sleepHandler calling tx_resume\n", NOW, index_);
	tx_resume();
}

void
Mac802_11M::nextWakeHandler()
{

	if(!nextWakeOpt)
		return;

	// We need to see which node to wake next
	// If there's a pktWait_, we should send it,
	// otherwise, if the packet we received has a final
	// IP destination and this node is not it, we need to
	// wake the next hop
	
	if(pktWait_) {
		lowtarget_->recv(pktWait_, (Handler*)0);
	}
	else if(final_ip_set) {
		// Need to lookup next hop at Routing Layer then
		// resolve the next hop IP with the ARP cache.
		// Then, create a dummy packet to pass to the low mac
		// lowtarget_->recv(pktWait_, (Handler*)0);
	}

}

void Mac802_11M::lowHandler(LowHandlerType t) {
	// Call function to send packet
	switch(t) {
		case LOW_DROP_PKT:

			if(netif_->is_on())
				netif_->node_sleep();

			if(pktWait_ == NULL) {
				pktWait_ = 0;
				return;
			}
			discard(pktWait_, DROP_MAC_LOW); pktWait_ = 0;
			// DO queue callback
	fprintf(stderr, "%.9f %d lowHandler calling tx_resume\n", NOW, index_);
			tx_resume();
#ifdef DEBUG
			fprintf(stderr, "%f: %d packet will be dropped\n", NOW, index_);
#endif
			break;
		case LOW_WAKE_TX:
#ifdef DEBUG
			fprintf(stderr, "%f: %d received wake tx signal\n", NOW, index_);
#endif
	
			if (!netif_->is_on()) {
				if(netif_->is_sleeping())
					netif_->node_wake();
				else if(netif_->is_off())
					netif_->node_on();

				if(pktWait_ != 0) {
					// Schedule packet t be sent when power state
					// transition is finished
					macTrans_.resched(netif_->getTransTime());
				}
				else {
					fprintf(stderr, "%f: %d no packet to send!\n", NOW, index_);
				}
			}
			break;
		case LOW_WAKE_RX:
#ifdef DEBUG
			fprintf(stderr, "%f: %d received wake rx signal\n", NOW, index_);
#endif

			if (!netif_->is_on()) {
				if(netif_->is_sleeping())
					netif_->node_wake();
				else if(netif_->is_off())
					netif_->node_on();

				fprintf(stderr, "%.9f %d lowHandler RX setting mhSleep\n", NOW, index_);
				mhSleep_.resched(MAC_MaxIdleTime);

			}

			break;
		default:
			fprintf(stderr, "%f: %d unknown LowHandlerType\n", NOW, index_);
			if(netif_->is_on())
				netif_->node_sleep();

	}
}
// end MATT

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac802_11MClass : public TclClass {
public:
	Mac802_11MClass() : TclClass("Mac/802_11M") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11M(&PMIB, &MMIB));
	}
} class_mac802_11m;


/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac802_11M::Mac802_11M(PHY_MIB *p, MAC_MIB *m) : Mac(), mhIF_(this), mhNav_(this), 
mhRecv_(this), mhSend_(this), mhDefer_(this, p->SlotTime), 
mhBackoff_(this, p->SlotTime),
// MATT
macTrans_(this), mhSleep_(this), mhNextWake_(this)
// end MATT
{
	macmib_ = m;
	phymib_ = p;
	
	nav_ = 0.0;
	
	tx_state_ = rx_state_ = MAC_IDLE;
	tx_active_ = 0;
	
	pktRTS_ = 0;
	pktCTRL_ = 0;

	// MATT
	pktWait_ = 0;
	rtsOpt = false;
	nextWakeOpt = false;
	final_ip_set = false;
	// end MATT

	cw_ = phymib_->CWMin;
	ssrc_ = slrc_ = 0;
	
	sifs_ = phymib_->SIFSTime;
	pifs_ = sifs_ + phymib_->SlotTime;
	difs_ = sifs_ + 2*phymib_->SlotTime;
	
	// see (802.11-1999, 9.2.10) 
	eifs_ = sifs_ + (8 * ETHER_ACK_LEN / phymib_->PLCPDataRate) + difs_;
	
	tx_sifs_ = sifs_ - phymib_->RxTxTurnaroundTime;
	tx_pifs_ = tx_sifs_ + phymib_->SlotTime;
	tx_difs_ = tx_sifs_ + 2 * phymib_->SlotTime;
	
	sta_seqno_ = 1;
	cache_ = 0;
	cache_node_count_ = 0;
	
	// chk if basic/data rates are set
	// otherwise use bandwidth_ as default;
	
	Tcl& tcl = Tcl::instance();
	tcl.evalf("Mac/802_11M set basicRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("basicRate_", &basicRate_);
	else
		basicRate_ = bandwidth_;

	tcl.evalf("Mac/802_11M set dataRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("dataRate_", &dataRate_);
	else
		dataRate_ = bandwidth_;
}


int
Mac802_11M::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} else if(strcmp(argv[1], "nodes") == 0) {
			if(cache_) return TCL_ERROR;
			cache_node_count_ = atoi(argv[2]);
			cache_ = new Host[cache_node_count_ + 1];
			assert(cache_);
			bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
			return TCL_OK;
		}
		// MATT
		else if(strcmp(argv[1], "low-target") == 0) {
			lowtarget_ = (MacLow *) TclObject::lookup(argv[2]);
			if(lowtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
		else if(strcmp(argv[1], "setRtsOpt") == 0) {
			if(strcmp(argv[2], "ON") == 0) {
				rtsOpt = true;
			}
			else {
				rtsOpt = false;
			}
			return TCL_OK;
		}
		else if(strcmp(argv[1], "setNextWakeOpt") == 0) {
			if(strcmp(argv[2], "ON") == 0) {
				nextWakeOpt = true;
			}
			else {
				nextWakeOpt = false;
			}
		}
		// end MATT
	}
	return Mac::command(argc, argv);
}

/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
Mac802_11M::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11m* dh = HDR_MAC802_11M(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
Mac802_11M::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %2.9f, idle: %d\n",
		tx_state_, rx_state_, nav_, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktRTS_: %x, pktCTRL_: %x, callback: %x\n",
		(int) pktTx_, (int) pktRx_, (int) pktRTS_,
		(int) pktCTRL_, (int) callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mhRecv_.busy(), mhSend_.busy(), mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff_.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline int
Mac802_11M::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac802_11m *dh = (struct hdr_mac802_11m*) hdr;
	//dst = (u_int32_t)(dst);

	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));

	return ETHER_ADDR(dh->dh_da);
}

inline int 
Mac802_11M::hdr_src(char* hdr, int src )
{
	struct hdr_mac802_11m *dh = (struct hdr_mac802_11m*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
	return ETHER_ADDR(dh->dh_sa);
}

inline int 
Mac802_11M::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac802_11m *dh = (struct hdr_mac802_11m*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */
inline int
Mac802_11M::is_idle()
{
	if(rx_state_ != MAC_IDLE)
		return 0;
	if(tx_state_ != MAC_IDLE)
		return 0;
	if(nav_ > Scheduler::instance().clock())
		return 0;
	
	return 1;
}

void
Mac802_11M::discard(Packet *p, const char* why)
{
	hdr_mac802_11m* mh = HDR_MAC802_11M(p);
	hdr_cmn *ch = HDR_CMN(p);

#if 0
	/* old logic 8/8/98 -dam */
	/*
	 * If received below the RXThreshold, then just free.
	 */
	if(p->txinfo_.Pr < p->txinfo_.ant.RXThresh) {
		Packet::free(p);
		//p = 0;
		return;
	}
#endif // 0

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		//p = 0;
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Management:
		drop(p, why);
		return;
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_RTS:
			if((u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			/* fall through - if necessary */
		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
			if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_sa) == \
			   (u_int32_t)index_ ||
			   (u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				drop(p);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
Mac802_11M::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */
	set_nav(usec(eifs_ + txtime(p)));
	Packet::free(p);
}

void
Mac802_11M::collision(Packet *p)
{
	switch(rx_state_) {
	case MAC_RECV:
		SET_RX_STATE(MAC_COLL);
		/* fall through */
	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(txtime(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(txtime(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
Mac802_11M::tx_resume()
{
	double rTime;
	assert(mhSend_.busy() == 0);
	assert(mhDefer_.busy() == 0);

	if(pktCTRL_) {
		/*
		 *  Need to send a CTS or ACK.
		 */
		fprintf(stderr, "%.9f %d calling defer SIFS pktCTRL_\n", NOW, index_);
		mhDefer_.start(sifs_);
	} else if(pktRTS_) {
		if(mhBackoff_.busy() == 0) {
			// MATT: Bug fix
			fprintf(stderr, "%.9f %d calling defer DIFS pktRTS_\n", NOW, index_);
			mhDefer_.start(difs_);
			// rTime = (Random::random() % cw_) * phymib_->SlotTime;
			// mhDefer_.start(difs_ + rTime);
			// end MATT
		}
	} else if(pktTx_) {
		if(mhBackoff_.busy() == 0) {
			hdr_cmn *ch = HDR_CMN(pktTx_);
			struct hdr_mac802_11m *mh = HDR_MAC802_11M(pktTx_);
			
			if ((u_int32_t) ch->size() < macmib_->RTSThreshold ||
			    (u_int32_t) ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				// MATT: Bug fix
				fprintf(stderr, "%.9f %d calling defer DIFS pktTx_\n", NOW, index_);
				mhDefer_.start(difs_);
				// rTime = (Random::random() % cw_) * phymib_->SlotTime;
				// mhDefer_.start(difs_ + rTime);
				// end MATT
			} else {
				fprintf(stderr, "%.9f %d calling defer SIFS pktTx_\n", NOW, index_);
				mhDefer_.start(sifs_);
			}
		}
	// MATT
	} else if(pktWait_) {

		// If we're using the optimization, the wakeup will be
		// done in nextWakeHandler
		if(nextWakeOpt && final_ip_set)
			return;

		// A packet arrived from the queue while we were on the receiving
		// end of another communication
		if(netif_->is_on()) {
			netif_->node_sleep();
			mhSleep_.force_cancel();
		}
#ifdef DEBUG
		fprintf(stderr, "%f: %d calling recv in pktWait_\n", NOW, index_);
#endif
		lowtarget_->recv(pktWait_, (Handler*)0);

	} else if(mhSleep_.status() == TIMER_IDLE && callback_) {
	// tx_resume gets called at the end of the timer
	// end MATT
#ifdef DEBUG
		fprintf(stderr, "%f: %d in callback\n", NOW, index_);
#endif
		Handler *h = callback_;
		callback_ = 0;
		// MATT
		// We need to set the tx state to idle before recv is
		// called by the queue
		SET_TX_STATE(MAC_IDLE);
		// end MATT
		h->handle((Event*) 0);
		// MATT
		if(netif_->is_on()) 
			netif_->node_sleep();	

		// end MATT
	}
	SET_TX_STATE(MAC_IDLE);
}

void
Mac802_11M::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	SET_RX_STATE(MAC_IDLE);
}

/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
void
Mac802_11M::backoffHandler()
{
	if(pktCTRL_) {
		assert(mhSend_.busy() || mhDefer_.busy());
		return;
	}

	if(check_pktRTS() == 0)
		return;

	fprintf(stderr,"%.9f %d calling check_pktTx in backoff\n", NOW, index_);
	if(check_pktTx() == 0) 
		return;
}

void
Mac802_11M::deferHandler()
{
	assert(pktCTRL_ || pktRTS_ || pktTx_);

	if(check_pktCTRL() == 0)
		return;

	assert(mhBackoff_.busy() == 0);
	//if (mhBackoff_.busy() != 0)
	//{
	//	printf("deferHandler:mhBackoff_ busy!\n");
	//	return;
	//}
	if(check_pktRTS() == 0)
		return;

	fprintf(stderr, "%.9f %d calling check_pktTx in defer\n", NOW, index_);
	if(check_pktTx() == 0)
		return;
}

void
Mac802_11M::navHandler()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(difs_);
}

void
Mac802_11M::recvHandler()
{
	recv_timer();
}

void
Mac802_11M::sendHandler()
{
	send_timer();
}


void
Mac802_11M::txHandler()
{
	tx_active_ = 0;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac802_11M::send_timer()
{
	fprintf(stderr, "%.9f %d calling send_timer\n", NOW, index_);
	switch(tx_state_) {
	/*
	 * Sent a RTS, but did not receive a CTS.
	 */
	case MAC_RTS:
		RetransmitRTS();
		break;
	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */
	case MAC_CTS:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		// MATT
		// Reschedule sleep time in case DATA packet
		// was lost due to collision
		fprintf(stderr, "%.9f %d send CTS setting mhSleep\n", NOW, index_);
		mhSleep_.resched(MAC_MaxIdleTime);
		// end MATT
		break;
	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */
	case MAC_SEND:
		RetransmitDATA();
		break;
	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */
	case MAC_ACK:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); pktCTRL_ = 0;
		// MATT
		// Give some time in case the ACK gets lost and the RTS
		// is resent
		if(netif_->is_on()) {
			// mhSleep_.resched(MAC_IdleAfterAck);
			netif_->node_sleep();
			mhSleep_.force_cancel();
		}
		// end MATT
		break;
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac802_11M::check_pktCTRL()
{
	struct hdr_mac802_11m *mh;
	double timeout;

	if(pktCTRL_ == 0)
		return -1;
	if(tx_state_ == MAC_CTS || tx_state_ == MAC_ACK)
		return -1;

	mh = HDR_MAC802_11M(pktCTRL_);
							  
	switch(mh->dh_fc.fc_subtype) {
	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */
	case MAC_Subtype_CTS:
		if(!is_idle()) {
			discard(pktCTRL_, DROP_MAC_BUSY); pktCTRL_ = 0;
			return 0;
		}
		SET_TX_STATE(MAC_CTS);
		
		/*
		 * timeout:  cts + data tx time calculated by
		 *           adding cts tx time to the cts duration
		 *           minus ack tx time -- this timeout is
		 *           a guess since it is unspecified
		 *           (note: mh->dh_duration == cf->cf_duration)
		 */
		timeout = txtime(ETHER_CTS_LEN, basicRate_)
			+ DSSS_MaxPropagationDelay			// XXX
			+ sec(mh->dh_duration)
			+ DSSS_MaxPropagationDelay			// XXX
			- sifs_
			- txtime(ETHER_ACK_LEN, basicRate_);
		
		break;
		/*
		 * IEEE 802.11 specs, section 9.2.8
		 * Acknowledments are sent after an SIFS, without regard to
		 * the busy/idle state of the medium.
		 */
	case MAC_Subtype_ACK:
		SET_TX_STATE(MAC_ACK);

		timeout = txtime(ETHER_ACK_LEN, basicRate_);
		
		break;
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
		exit(1);
	}
        TRANSMIT(pktCTRL_, timeout);
	return 0;
}

int
Mac802_11M::check_pktRTS()
{
	struct hdr_mac802_11m *mh;
	double timeout;
	// MATT
	double rts_tx_time;
	if(nextWakeOpt)
		rts_tx_time = txtime(ETHER_RTS_IP_LEN, basicRate_);
	else
		rts_tx_time = txtime(ETHER_RTS_LEN, basicRate_);
	// end MATT

	assert(mhBackoff_.busy() == 0);

	if(pktRTS_ == 0)
 		return -1;
	//struct hdr_cmn *ch = HDR_CMN(pktRTS_);
	mh = HDR_MAC802_11M(pktRTS_);

 	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_RTS:
		if(! is_idle()) {
			inc_cw();
			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		SET_TX_STATE(MAC_RTS);
		// MATT
		// timeout = txtime(ETHER_RTS_LEN, basicRate_)
		timeout = rts_tx_time
		// end MATT
			+ DSSS_MaxPropagationDelay			// XXX
			+ sifs_
			+ txtime(ETHER_CTS_LEN, basicRate_)
			+ DSSS_MaxPropagationDelay;			// XXX
		break;
	default:
		fprintf(stderr, "%f %d check_pktRTS:Invalid MAC Control subtype\n", 
							 NOW, index_);
		exit(1);
	}
        TRANSMIT(pktRTS_, timeout);
	return 0;
}

int
Mac802_11M::check_pktTx()
{
	struct hdr_mac802_11m *mh;
	double timeout;
	
	assert(mhBackoff_.busy() == 0);

	if(pktTx_ == 0)
		return -1;

	mh = HDR_MAC802_11M(pktTx_);
       	//int len = HDR_CMN(pktTx_)->size();

	fprintf(stderr, "%.9f %d in check_pktTx\n", NOW, index_);
	trace_pkt(pktTx_);

	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {
			sendRTS(ETHER_ADDR(mh->dh_da));
			inc_cw();
			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		SET_TX_STATE(MAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_da) != MAC_BROADCAST)
			timeout = txtime(pktTx_)
				+ DSSS_MaxPropagationDelay		// XXX
				+ sifs_
				+ txtime(ETHER_ACK_LEN, basicRate_)
				+ DSSS_MaxPropagationDelay;		// XXX
		else
			timeout = txtime(pktTx_);
		break;
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
		//printf("pktRTS:%x, pktCTS/ACK:%x, pktTx:%x\n",pktRTS_, pktCTRL_,pktTx_);
		exit(1);
	}
        TRANSMIT(pktTx_, timeout);
	return 0;
}

/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
void
Mac802_11M::sendRTS(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);
	
	assert(pktTx_);
	assert(pktRTS_ == 0);

	// MATT
	// Get IP address for next hop off of pktTx_ if doing optimization
	nsaddr_t final_ip;
	struct hdr_ip *ih = HDR_IP(pktTx_);
	if(nextWakeOpt)
		final_ip = ih->daddr();
	// end MATT

	/*
	 *  If the size of the packet is larger than the
	 *  RTSThreshold, then perform the RTS/CTS exchange.
	 *
	 *  XXX: also skip if destination is a broadcast
	 */
	if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_->RTSThreshold ||
	    (u_int32_t) dst == MAC_BROADCAST) {
		Packet::free(p);
		//p = 0;
		return;
	}

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	// MATT
	// ch->size() = ETHER_RTS_LEN;
	if(nextWakeOpt)
		ch->size() = ETHER_RTS_IP_LEN;
	else
		ch->size() = ETHER_RTS_LEN;
	// end MATT
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rf->rf_fc.fc_type	= MAC_Type_Control;
 	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
 	rf->rf_fc.fc_to_ds	= 0;
 	rf->rf_fc.fc_from_ds	= 0;
 	rf->rf_fc.fc_more_frag	= 0;
 	rf->rf_fc.fc_retry	= 0;
 	rf->rf_fc.fc_pwr_mgt	= 0;
 	rf->rf_fc.fc_more_data	= 0;
 	rf->rf_fc.fc_wep	= 0;
 	rf->rf_fc.fc_order	= 0;

	//rf->rf_duration = RTS_DURATION(pktTx_);
	STORE4BYTE(&dst, (rf->rf_ra));
	
	/* store rts tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	STORE4BYTE(&index_, (rf->rf_ta));
	/* calculate rts duration field */
	rf->rf_duration = usec(sifs_
			       + txtime(ETHER_CTS_LEN, basicRate_)
			       + sifs_
			       + txtime(pktTx_)
			       + sifs_
			       + txtime(ETHER_ACK_LEN, basicRate_));
	
	// MATT
	if(nextWakeOpt)
		rf->final_ip = final_ip;
	// end MATT
	
	pktRTS_ = p;
}

void
Mac802_11M::sendCTS(int dst, double rts_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_CTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	//ch->direction() = hdr_cmn::DOWN;
	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;
	
	//cf->cf_duration = CTS_DURATION(rts_duration);
	STORE4BYTE(&dst, (cf->cf_ra));
	
	/* store cts tx time */
	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate cts duration */
	cf->cf_duration = usec(sec(rts_duration)
			       - sifs_
			       - txtime(ETHER_CTS_LEN, basicRate_));
	
	pktCTRL_ = p;
	
}

void
Mac802_11M::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	//af->af_duration = ACK_DURATION();
	STORE4BYTE(&dst, (af->af_ra));

	/* store ack tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate ack duration */
 	af->af_duration = 0;	
	
	pktCTRL_ = p;
}

void
Mac802_11M::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11m* dh = HDR_MAC802_11M(p);

	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += ETHER_HDR_LEN11;

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	//printf(".....p = %x, mac-subtype-%d\n",p,dh->dh_fc.fc_subtype);
	
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	/* store data tx time */
 	ch->txtime() = txtime(ch->size(), dataRate_);

	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST) {
		/* store data tx time for unicast packets */
		ch->txtime() = txtime(ch->size(), dataRate_);
		
		//dh->dh_duration = DATA_DURATION();
		dh->dh_duration = usec(txtime(ETHER_ACK_LEN, basicRate_)
 				       + sifs_);
	} else {
		/* store data tx time for broadcast packets (see 9.6) */
		ch->txtime() = txtime(ch->size(), basicRate_);
		
		dh->dh_duration = 0;
	}
	pktTx_ = p;
}

/* ======================================================================
   Retransmission Routines
   ====================================================================== */
void
Mac802_11M::RetransmitRTS()
{
	assert(pktTx_);
	assert(pktRTS_);
	assert(mhBackoff_.busy() == 0);

	macmib_->RTSFailureCount++;

	ssrc_ += 1;			// STA Short Retry Count

	if(ssrc_ >= macmib_->ShortRetryLimit) {
		discard(pktRTS_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS_ = 0;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        /*
                         *  Need to remove the MAC header so that 
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */
                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
		//printf("(%d)....discarding RTS:%x\n",index_,pktRTS_);
		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		ssrc_ = 0;
		rst_cw();
	} else {
		//printf("(%d)...retxing RTS:%x\n",index_,pktRTS_);
		struct rts_frame *rf;
		rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);
		rf->rf_fc.fc_retry = 1;

		inc_cw();
		mhBackoff_.start(cw_, is_idle());
	}
}

void
Mac802_11M::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac802_11m *mh;
	u_int32_t *rcount, *thresh;

	assert(mhBackoff_.busy() == 0);

	assert(pktTx_);
	assert(pktRTS_ == 0);

	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC802_11M(pktTx_);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if((u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
		Packet::free(pktTx_); pktTx_ = 0;
		// MATT
		pktWait_ = 0;
		// end MATT

		/*
		 * Backoff at end of TX.
		 */
		rst_cw();
		mhBackoff_.start(cw_, is_idle());

		return;
	}

	macmib_->ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib_->RTSThreshold) {
		rcount = &ssrc_;
		thresh = &macmib_->ShortRetryLimit;
	}
	else {
		rcount = &slrc_;
		thresh = &macmib_->LongRetryLimit;
	}

	(*rcount)++;

	if(*rcount > *thresh) {
		macmib_->FailedCount++;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx_ = 0;
		//printf("(%d)DATA discarded: count exceeded\n",index_);
		*rcount = 0;
		rst_cw();
	}
	else {
		struct hdr_mac802_11m *dh;
		dh = HDR_MAC802_11M(pktTx_);
		dh->dh_fc.fc_retry = 1;

		sendRTS(ETHER_ADDR(mh->dh_da));
		//printf("(%d)retxing data:%x..sendRTS..\n",index_,pktTx_);
		inc_cw();
		mhBackoff_.start(cw_, is_idle());
	}
}

/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac802_11M::send(Packet *p, Handler *h)
{
	double rTime;
	struct hdr_mac802_11m* dh = HDR_MAC802_11M(p);

	/* 
	 * drop the packet if the node is in sleep mode
	 XXX sleep mode can't stop node from sending packets
	 */
	// MATT
	/*
	EnergyModel *em = netif_->node()->energy_model();
	if (em && em->sleep()) {
		em->set_node_sleep(0);
		em->set_node_state(EnergyModel::INROUTE);
	}
	*/
	if(netif_->is_sleeping() || netif_->is_off()) {
		fprintf(stderr, "Trying to send while asleep or off\n");
	}
	// end MATT
	
	// MATT
	// This is now done in recv
	// callback_ = h;
	// end MATT
	sendDATA(p);
	// MATT
	// pktTx_ now points to the packet
	pktWait_ = 0;	
	// end MATT
	sendRTS(ETHER_ADDR(dh->dh_da));

	/*
	 * Assign the data packet a sequence number.
	 */
	dh->dh_scontrol = sta_seqno_++;

	/*
	 *  If the medium is IDLE, we must wait for a DIFS
	 *  Space before transmitting.
	 */
	if(mhBackoff_.busy() == 0) {
		if(is_idle()) {
			/*
			 * If we are already deferring, there is no
			 * need to reset the Defer timer.
			 */
			if(mhDefer_.busy() == 0) {
				// MATT: Bug fix
				if(!rtsOpt || ETHER_ADDR(dh->dh_da) == MAC_BROADCAST) {
				fprintf(stderr, "%.9f %d calling defer DIFS recv\n", NOW, index_);
				mhDefer_.start(difs_);
				// rTime = (Random::random() % cw_) * (phymib_->SlotTime);
				// mhDefer_.start(difs_ + rTime);
				}
				else {

					// Spoof an RTS having been sent and the CTS
					// having been received.  I.e., let the WAKE and
					// WAKE-ACK serve that purpose
					double rts_tx_time;
					if(nextWakeOpt)
						rts_tx_time = txtime(ETHER_RTS_IP_LEN, basicRate_);
					else
						rts_tx_time = txtime(ETHER_RTS_LEN, basicRate_);

					double t  = rts_tx_time
						+ DSSS_MaxPropagationDelay			// XXX
						+ sifs_
						+ txtime(ETHER_CTS_LEN, basicRate_)
						+ DSSS_MaxPropagationDelay;			// XXX
					mhSend_.start(t);
					SET_TX_STATE(MAC_RTS);

					Packet *p = Packet::alloc();
					hdr_cmn* ch = HDR_CMN(p);
					struct cts_frame *cf = 
							(struct cts_frame*)p->access(hdr_mac::offset_);
					ch->uid() = 0;
					ch->ptype() = PT_MAC;
					ch->size() = ETHER_CTS_LEN;
					ch->iface() = -2;
					ch->error() = 0;
					bzero(cf, MAC_HDR_LEN);
					cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
					cf->cf_fc.fc_type	= MAC_Type_Control;
					cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
				 	cf->cf_fc.fc_to_ds	= 0;
				 	cf->cf_fc.fc_from_ds	= 0;
				 	cf->cf_fc.fc_more_frag	= 0;
				 	cf->cf_fc.fc_retry	= 0;
				 	cf->cf_fc.fc_pwr_mgt	= 0;
				 	cf->cf_fc.fc_more_data	= 0;
				 	cf->cf_fc.fc_wep	= 0;
				 	cf->cf_fc.fc_order	= 0;
					STORE4BYTE(&index_, (cf->cf_ra));
					ch->txtime() = txtime(ch->size(), basicRate_);
					/* zero curations in CTS distinguish this in the logs */
					cf->cf_duration = 0;
	
					recvCTS(p);

				}  // end rtsOpt
				// end MATT
			}
			
		}
	/*
	 * If the medium is NOT IDLE, then we start
	 * the backoff timer.
	 */
		else {
			mhBackoff_.start(cw_, is_idle());
		}
	}
}

void
Mac802_11M::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	fprintf(stderr, "%f: %d Receiving packet\n", NOW, index_);
	if(hdr->direction() == hdr_cmn::DOWN) {
		// MATT
		fprintf(stderr, "%f: %d Receiving packet from queue\n", NOW, index_);
		if(netif_->is_on()) {
#ifdef DEBUG
			fprintf(stderr, "%f: %d Receiving packet while on!!!\n", NOW, index_);
			trace_pkt(p);
#endif
		}

		if(pktWait_ != 0)
			fprintf(stderr, "%f %d pktWait_ is not 0\n", NOW, index_);
		if(callback_ != 0)
			fprintf(stderr, "%f %d callback_ is not 0\n", NOW, index_);

		// Stop the sleep timer if it is ticking
		mhSleep_.force_cancel();

		pktWait_ = p;
		callback_ = h;
		if(is_idle()) {
			lowtarget_->recv(p, (Handler*)0);
		}
		else {
#ifdef DEBUG
			fprintf(stderr, "%.9f %d: not idle\n", NOW, index_);
			if(rx_state_ != MAC_IDLE)
			fprintf(stderr, "\trx not idle\n");
			if(tx_state_ != MAC_IDLE)
			fprintf(stderr, "\ttx not idle\n");
			if(nav_ > NOW)
			fprintf(stderr, "\tnav not idle\n");
#endif
		}
		/*
		fprintf(stderr, "netif_: %X\n", netif_);
		if(netif_->is_on())
			fprintf(stderr, "Receiving packet while on!!!\n");
		else
			fprintf(stderr, "Receiving packet while off\n");
		*/
		// send(p, h);
		// end MATT
		return;
	}
	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);
		pktRx_ = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(txtime(p));

		// MATT
		// If we have a next hop for the optimization, start another
		// timer
		// Assume this is a DATA packet
		if(nextWakeOpt) {
			double nextWakeTime = txtime(p) - lowtarget_->getWakeDelay();
			mhNextWake_.resched(nextWakeTime);
		}
		// end MATT
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
		} else {
			collision(p);
		}
	}
}

void
Mac802_11M::recv_timer()
{
	u_int32_t src; 
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac802_11m *mh = HDR_MAC802_11M(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	// XXX debug
	//struct cts_frame *cf = (struct cts_frame*)pktRx_->access(hdr_mac::offset_);
	//u_int32_t src = ETHER_ADDR(mh->dh_sa);
	
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	// MATT
	bool pktRxFreed = false;
	// end MATT

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);
	
	trace_pkt(pktRx_);
        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active_) {
                Packet::free(pktRx_);
					 // MATT
					 pktRxFreed = true;
					 // end MATT
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == MAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);
		set_nav(usec(eifs_));
		// MATT
		pktRxFreed = true;
		// end MATT
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx_);
		set_nav(usec(eifs_));
		// MATT
		pktRxFreed = true;
		// end MATT
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
		set_nav(mh->dh_duration);
	}

        /* tap out - */
        if (tap_ && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype ) 
		tap_->tap(pktRx_);
	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation 
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */
	if (netif_->node()->energy_model() && 
	    netif_->node()->energy_model()->adaptivefidelity()) {
		src = ETHER_ADDR(mh->dh_sa);
		netif_->node()->energy_model()->add_neighbor(src);
	}
	/*
	 * Address Filtering
	 */
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {

	case MAC_Type_Management:
		discard(pktRx_, DROP_MAC_PACKET_ERROR);
		goto done;
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
			recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
			recvCTS(pktRx_);
			break;
		case MAC_Subtype_ACK:
			recvACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(subtype) {
		case MAC_Subtype_Data:
			recvDATA(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
 done:
	trace_pkt(pktRx_);
	pktRx_ = 0;
	rx_resume();

	// MATT
	if(nextWakeOpt) {
		if(final_ip_set)
			final_ip_set = false;
		mhNextWake_.force_cancel();
	}
	// end MATT

	// MATT
	// The following sequence can happen causing deadlock:
	// 1. recvACK gets called, this calls tx_resume
	// 2. in tx_resume, queue callback happens
	// 3. queue handler calls MAC's recv function to pass new packet
	// 4. we call is_idle before passing the packet to low-target
	// 5. is_idle is false since tx_state will only be set to MAC_IDLE
	// after MAC recv and queue handle complete and rx_state will only be
	// set to idle in rx_resume above.
	// 6. Therefore, there after receiving an ACK here, there could be
	// a pktWait_ that gets held indefinitely until tx_resume is called
	// again.
	if(!pktRxFreed && subtype == MAC_Subtype_ACK) {
	fprintf(stderr, "%.9f %d recv_timer calling tx_resume\n", NOW, index_);
		tx_resume();
	}
	// end MATT
}


void
Mac802_11M::recvRTS(Packet *p)
{
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	if(tx_state_ != MAC_IDLE) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this RTS.
	 */
	if(pktCTRL_) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);

	/*
	 *  Stop deferring - will be reset in tx_resume().
	 */
	fprintf(stderr, "%.9f %d calling defer stop recvRTS\n", NOW, index_);
	if(mhDefer_.busy()) mhDefer_.stop();


	// MATT
	if(nextWakeOpt) {
		cur_final_ip = rf->final_ip;
		final_ip_set = true;
	}
	// end MATT

	tx_resume();

	mac_log(p);
}

/*
 * txtime()	- pluck the precomputed tx time from the packet header
 */
double
Mac802_11M::txtime(Packet *p)
 {
	 struct hdr_cmn *ch = HDR_CMN(p);
	 double t = ch->txtime();
	 if (t < 0.0) {
		 drop(p, "XXX");
 		exit(1);
	 }
	 return t;
 }

 
/*
 * txtime()	- calculate tx time for packet of size "psz" bytes 
 *		  at rate "drt" bps
 */
double
Mac802_11M::txtime(double psz, double drt)
{
	double dsz = psz - PLCP_HDR_LEN;
	int plcp_hdr = PLCP_HDR_LEN << 3;
	int datalen = (int)dsz << 3;
	
	double t = (((double)plcp_hdr)/phymib_->PLCPDataRate) + (((double)datalen)/drt);
	return(t);
}



void
Mac802_11M::recvCTS(Packet *p)
{
	if(tx_state_ != MAC_RTS) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktRTS_);
	Packet::free(pktRTS_); pktRTS_ = 0;

	assert(pktTx_);
	// debug
	//struct hdr_mac802_11 *mh = HDR_MAC802_11(pktTx_);
	//printf("(%d):recvCTS:pktTx_-%x,mac-subtype-%d & pktCTS_:%x\n",index_,pktTx_,mh->dh_fc.fc_subtype,p);
	
	mhSend_.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */
	//ssrc_ = 0;
	//rst_cw();

	tx_resume();

	mac_log(p);
}

void
Mac802_11M::recvDATA(Packet *p)
{
	struct hdr_mac802_11m *dh = HDR_MAC802_11M(p);
	u_int32_t dst, src, size;

	{	struct hdr_cmn *ch = HDR_CMN(p);

		dst = ETHER_ADDR(dh->dh_da);
		src = ETHER_ADDR(dh->dh_sa);
		size = ch->size();

		/*
		 * Adjust the MAC packet size - ie; strip
		 * off the mac header
		 */
		ch->size() -= ETHER_HDR_LEN11;
		ch->num_forwards() += 1;
	}

	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
		if(size >= macmib_->RTSThreshold) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();
				/*
				 * Our CTS got through.
				 */
				//printf("(%d): RECVING DATA!\n",index_);
				//ssrc_ = 0;
				//rst_cw();
			}
			// MATT
			// else {
			// pretend we send a CTS for this packet
			else if(!rtsOpt) {
			// end MATT
				discard(p, DROP_MAC_BUSY);
				//printf("(%d)..discard DATA\n",index_);
				return;
			}
			sendACK(src);
			tx_resume();
		}
		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */
		else {
			if(pktCTRL_) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend_.busy() == 0)
				tx_resume();
		}
	}

	/* ============================================================
	    Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was 
	   suggested by Joerg Diederich <dieder@@ibr.cs.tu-bs.de>. 
	   Changed on 19th Oct'2000 */

        if(dst != MAC_BROADCAST) {
                if (src < (u_int32_t) cache_node_count_) {
                        Host *h = &cache_[src];

                        if(h->seqno && h->seqno == dh->dh_scontrol) {
                                discard(p, DROP_MAC_DUPLICATE);
                                return;
                        }
                        h->seqno = dh->dh_scontrol;
                } else {
			static int count = 0;
			if (++count <= 10) {
				printf ("MAC_802_11M: accessing MAC cache_ array out of range (src %u, dst %u, size %d)!\n", src, dst, cache_node_count_);
				if (count == 10)
					printf ("[suppressing additional MAC cache_ warnings]\n");
			};
		};
	}

	/*

	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	//p->incoming = 1;
	// XXXXX NOTE: use of incoming flag has been depracated; In order to track direction of pkt flow, direction_ in hdr_cmn is used instead. see packet.h for details. 
	
	// MATT
	if(dst == MAC_BROADCAST) {
		if(netif_->is_on()) {
			netif_->node_sleep();
			mhSleep_.force_cancel();
		}
		// Need to make sure we check the pktWait_ if we
		// get a broadcast packet
	fprintf(stderr, "%.9f %d recvDATA calling tx_resume\n", NOW, index_);
		tx_resume();
	}
	// end MATT
	
	uptarget_->recv(p, (Handler*) 0);
}


void
Mac802_11M::recvACK(Packet *p)
{
	
	struct hdr_cmn *ch = HDR_CMN(p);

	if(tx_state_ != MAC_SEND) {
	discard(p, DROP_MAC_INVALID_STATE);
	return;
	}
	//printf("(%d)...................recving ACK:%x\n",index_,p);
	assert(pktTx_);
	Packet::free(pktTx_); pktTx_ = 0;
	// MATT
	pktWait_ = 0;
	// end MATT

	mhSend_.stop();

	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */
	if((u_int32_t) ch->size() <= macmib_->RTSThreshold)
		ssrc_ = 0;
	else
		slrc_ = 0;

	/*
	 * Backoff before sending again.
	 */
	rst_cw();
	assert(mhBackoff_.busy() == 0);
	mhBackoff_.start(cw_, is_idle());

	tx_resume();

	mac_log(p);
}



@
