/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * $Header: /home/mobiserve2/mjmille2/CVS_REPOSITORY/ns-2.26/mac/Attic/mac-low.cc,v 1.1.2.13 2003/04/30 22:51:50 mjmille2 Exp $
 *
 */

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

 #define DEBUG 99

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-low-timers.h"
#include "mac-low.h"
#include "cmu-trace.h"
#include "mac-802_11m.h"

// XXX Can't we make these macros inline methods? Otherwise why should we have
// inline methods at all??

#define CHECK_BACKOFF_TIMER()						\
{									\
	if(is_idle() && mhBackoff_.paused())				\
		/* mhBackoff_.resume(difs_); */		\
		/* Don't defer */ \
		mhBackoff_.resume(0.0);				\
	if(! is_idle() && mhBackoff_.busy() && ! mhBackoff_.paused())	\
		mhBackoff_.pause();					\
}

// Checked
#define TRANSMIT(p, t)                                                  \
{                                                                       \
	tx_active_ = 1;                                                  \
                                                                        \
        /*                                                              \
         * If I'm transmitting without doing CS, such as when           \
         * sending an ACK, any incoming packet will be "missed"         \
         * and hence, must be discarded.                                \
         */                                                             \
        if(rx_state_ != MAC_IDLE) {                                      \
                struct hdr_mac_low *dh = HDR_MAC_LOW(p);                  \
                                                                        \
                assert(pktRx_);                                          \
                struct hdr_cmn *ch = HDR_CMN(pktRx_);                    \
                                                                        \
                ch->error() = 1;        /* force packet discard */      \
        }                                                               \
                                                                        \
        /*                                                              \
         * pass the packet on the "interface" which will in turn        \
         * place the packet on the channel.                             \
         *                                                              \
         * NOTE: a handler is passed along so that the Network          \
         *       Interface can distinguish between incoming and         \
         *       outgoing packets.                                      \
         */                                                             \
        downtarget_->recv(p->copy(), this);                             \
                                                                        \
        mhSend_.start(t);                                                \
                                                                        \
	mhIF_.start(txtime(p));                                         \
}

#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);		\
	CHECK_BACKOFF_TIMER();		\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);			\
	CHECK_BACKOFF_TIMER();			\
}

/* ======================================================================
   Global Variables
   ====================================================================== */

static PHY_LOW_MIB PMIB =
{
	DSSS_LOW_CWMin, DSSS_LOW_CWMax, DSSS_LOW_SlotTime, DSSS_LOW_CCATime,
	DSSS_LOW_RxTxTurnaroundTime, DSSS_LOW_SIFSTime, DSSS_LOW_PreambleLength,
	DSSS_LOW_PLCPHeaderLength, DSSS_LOW_PLCPDataRate
};	

static MAC_LOW_MIB MMIB =
{
	MAC_LOW_RTSThreshold, MAC_LOW_WakeRetryLimit,
	MAC_LOW_LongRetryLimit, MAC_LOW_FragmentationThreshold,
	MAC_LOW_MaxTransmitMSDULifetime, MAC_LOW_MaxReceiveLifetime,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class MacLowClass : public TclClass {
public:
	MacLowClass() : TclClass("Mac/MacLow") {}
	TclObject* create(int, const char*const*) {
		return (new MacLow(&PMIB, &MMIB));
	}
} class_maclow;


/* ======================================================================
   Mac Class Functions
   ====================================================================== */
MacLow::MacLow(PHY_LOW_MIB *p, MAC_LOW_MIB *m) : Mac(), mhIF_(this), mhNav_(this), mhRecv_(this), mhSend_(this), mhDefer_(this, p->SlotTime), mhBackoff_(this, p->SlotTime)
{

	macLowCorrection();

	macmib_ = m;
	phymib_ = p;
	
	nav_ = 0.0;
	
	tx_state_ = rx_state_ = MAC_IDLE;
	tx_active_ = 0;
	
	pktWAKE_ = 0;
	pktWAKE_ACK_ = 0;

	cw_ = phymib_->CWMin;
	ssrc_ = 0;
	
	/*
	sifs_ = phymib_->SIFSTime;
	pifs_ = sifs_ + phymib_->SlotTime;
	difs_ = sifs_ + 2*phymib_->SlotTime;
	
	// see (802.11-1999, 9.2.10) 
	eifs_ = sifs_ + (8 * ETHER_ACK_LEN / phymib_->PLCPDataRate) + difs_;
	
	tx_sifs_ = sifs_ - phymib_->RxTxTurnaroundTime;
	tx_pifs_ = tx_sifs_ + phymib_->SlotTime;
	tx_difs_ = tx_sifs_ + 2 * phymib_->SlotTime;
	*/
	
	sta_seqno_ = 1;
	cache_ = 0;
	cache_node_count_ = 0;

	primaryMAC_ = 0;

	rtsOpt = false;
	nextWakeOpt = false;
	
	// chk if basic/data rates are set
	// otherwise use bandwidth_ as default;
	
	Tcl& tcl = Tcl::instance();
	tcl.evalf("Mac/MacLow set basicRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("basicRate_", &basicRate_);
	else 
		basicRate_ = bandwidth_;
	
	tcl.evalf("Mac/MacLow set dataRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("dataRate_", &dataRate_);
	else 
		dataRate_ = bandwidth_;
	
	/*
	fprintf(stderr, "WAKE LEN: %d\n", ETHER_WAKE_LEN);
	fprintf(stderr, "WAKE ACK LEN: %d\n", ETHER_WAKE_ACK_LEN);
	fprintf(stderr, "PLCP LEN: %d\n", PLCP_LOW_HDR_LEN);
	fprintf(stderr, "CONTROL LEN: %d\n", sizeof(struct frame_control_low));
	*/
}


int
MacLow::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} else if(strcmp(argv[1], "nodes") == 0) {
			if(cache_) return TCL_ERROR;
			cache_node_count_ = atoi(argv[2]);
			cache_ = new Host[cache_node_count_ + 1];
			assert(cache_);
			bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
			return TCL_OK;
		}
		else if(strcmp(argv[1], "primaryMAC") == 0) {
			TclObject *obj = (NsObject*) TclObject::lookup(argv[2]);
			if(obj == 0) { 
				primaryMAC_ = 0;
				return TCL_ERROR;
			}
			else {
				primaryMAC_ = (Mac802_11M*) obj;
				return TCL_OK;
			}
		}
		else if(strcmp(argv[1], "setRtsOpt") == 0) {
			if(strcmp(argv[2], "ON") == 0) {
				rtsOpt = true;
			}
			else {
				rtsOpt = false;
			}
			return TCL_OK;
		}
		else if(strcmp(argv[1], "setNextWakeOpt") == 0) {
			if(strcmp(argv[2], "ON") == 0) {
				nextWakeOpt = true;
			}
			else {
				nextWakeOpt = false;
			}
		}
	}
	return Mac::command(argc, argv);
}

/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
MacLow::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_low* dh = HDR_MAC_LOW(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
MacLow::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %2.9f, idle: %d\n",
		tx_state_, rx_state_, nav_, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktWAKE_: %x, pktWAKE_ACK_: %x, callback: %x\n",
		(int) pktTx_, (int) pktRx_, (int) pktWAKE_,
		(int) pktWAKE_ACK_, (int) callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mhRecv_.busy(), mhSend_.busy(), mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff_.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline int
MacLow::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac_low *dh = (struct hdr_mac_low*) hdr;
	//dst = (u_int32_t)(dst);

	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));

	return ETHER_ADDR(dh->dh_da);
}

inline int 
MacLow::hdr_src(char* hdr, int src )
{
	struct hdr_mac_low *dh = (struct hdr_mac_low*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
	return ETHER_ADDR(dh->dh_sa);
}

inline int 
MacLow::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac_low *dh = (struct hdr_mac_low*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */

double
MacLow::getWakeDelay()
{
	double ret_val = txtime(ETHER_WAKE_LEN, basicRate_)
		+ DSSS_LOW_MaxPropagationDelay			// XXX
		// XXX: New turnaround time here?
		// + sifs_
		+ DSSS_RxTxTurnaroundTime
		+ txtime(ETHER_WAKE_ACK_LEN, basicRate_)
		+ DSSS_LOW_MaxPropagationDelay			// XXX
		// Add a little extra time to account for congestion possibility
		+ cw_; 

	return ret_val;

}

inline int
MacLow::is_idle()
{
	if(rx_state_ != MAC_IDLE)
		return 0;
	if(tx_state_ != MAC_IDLE)
		return 0;
	if(nav_ > NOW)
		return 0;
	
	return 1;
}

// Checked
void
MacLow::discard(Packet *p, const char* why)
{
	hdr_mac_low* mh = HDR_MAC_LOW(p);
	hdr_cmn *ch = HDR_CMN(p);

	// if the rcvd pkt contains errors, a real MAC layer couldn't
	//   necessarily read any data from it, so we just toss it now 
	if(ch->error() != 0) {
		Packet::free(p);
		//p = 0;
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_LOW_Type_Wake: 
		if((u_int32_t)ETHER_ADDR(mh->dh_da) == \
		   (u_int32_t)index_ ||
		   (u_int32_t)ETHER_ADDR(mh->dh_sa) == \
		   (u_int32_t)index_ ||
		   (u_int32_t)ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
			drop(p, why);
			return;
		}
		break;
	case MAC_LOW_Type_WakeAck:
		if((u_int32_t)ETHER_ADDR(mh->dh_da) == (u_int32_t)index_ ) {
			drop(p, why);
			return;
		}
		break;
	default:
		fprintf(stderr, "invalid MAC LOW type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
MacLow::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */
	//set_nav(usec(eifs_ + txtime(p)));
	set_nav(usec(txtime(p)));
	Packet::free(p);
}

// Checked
void
MacLow::collision(Packet *p)
{
	switch(rx_state_) {
	case MAC_RECV:
		SET_RX_STATE(MAC_COLL);
		/* fall through */
	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(txtime(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(txtime(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
MacLow::callbackPrimary(LowHandlerType lt)
{
	if(primaryMAC_) {
		primaryMAC_->lowHandler(lt);
	}
}

void
MacLow::tx_resume()
{
	double rTime;
	assert(mhSend_.busy() == 0);
	assert(mhDefer_.busy() == 0);

	if(pktWAKE_ACK_) {
		/*
		if(mhBackoff_.busy() == 0) {
			fprintf(stderr, "%f %d calling check_pktWAKE_ACK\n", NOW, index_);
			check_pktWAKE_ACK();
			return;
		// mhDefer_.start(sifs_);
		}
		*/
	} else if(pktWAKE_) {
		/*
		if(mhBackoff_.busy() == 0) {
			check_pktWAKE();
			return;
		}
		*/
	} else if(pktTx_) {
		// pktTx_ = 0;
	}

	SET_TX_STATE(MAC_IDLE);
}

void
MacLow::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	SET_RX_STATE(MAC_IDLE);
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
// Checked
void
MacLow::backoffHandler()
{
	if(check_pktWAKE_ACK() == 0) 		
		return;

	if(check_pktWAKE() == 0)
		return;

}

void
MacLow::deferHandler()
{

	if(check_pktWAKE_ACK() == 0) 		
		return;

	if(check_pktWAKE() == 0)
		return;

/*
	assert(pktCTRL_ || pktRTS_ || pktTx_);

	if(check_pktCTRL() == 0)
		return;

	assert(mhBackoff_.busy() == 0);
	//if (mhBackoff_.busy() != 0)
	//{
	//	printf("deferHandler:mhBackoff_ busy!\n");
	//	return;
	//}
	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
*/
}

// Checked
void
MacLow::navHandler()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(0.0);
}

void
MacLow::recvHandler()
{
	recv_timer();
}

void
MacLow::sendHandler()
{
	send_timer();
}


void
MacLow::txHandler()
{
	tx_active_ = 0;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
// Checked
void
MacLow::send_timer()
{
	switch(tx_state_) {
	/*
	 * Sent a WAKE, didn't recv WAKE-ACK
	 */
	case MAC_WAKE:
		RetransmitWAKE();
		break;
	/*
	 * Sent a WAKE-ACK
	 */
	case MAC_WAKE_ACK:
		assert(pktWAKE_ACK_);
		Packet::free(pktWAKE_ACK_); pktWAKE_ACK_ = 0;
		callbackPrimary(LOW_WAKE_RX);
		break;
	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
// Checked
int
MacLow::check_pktWAKE_ACK()
{
	struct hdr_mac_low *mh;
	double timeout;

	if(pktWAKE_ACK_ == 0)
		return -1;
	if(tx_state_ == MAC_WAKE_ACK)
		return -1;

	mh = HDR_MAC_LOW(pktWAKE_ACK_);
							  
	switch(mh->dh_fc.fc_type) {
	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */
	case MAC_LOW_Type_WakeAck:
		if(!is_idle()) {
			fprintf(stderr, "%f %d dropping in check_pktWAKE_ACK\n", NOW, index_);
			discard(pktWAKE_ACK_, DROP_MAC_BUSY); pktWAKE_ACK_ = 0;
			return 0;
		}
		SET_TX_STATE(MAC_WAKE_ACK);
		
		/*
		 * timeout:  cts + data tx time calculated by
		 *           adding cts tx time to the cts duration
		 *           minus ack tx time -- this timeout is
		 *           a guess since it is unspecified
		 *           (note: mh->dh_duration == cf->cf_duration)
		 */
		/*
		timeout = txtime(ETHER_CTS_LEN, basicRate_)
			+ DSSS_LOW_MaxPropagationDelay			// XXX
			+ sec(mh->dh_duration)
			+ DSSS_LOW_MaxPropagationDelay			// XXX
			- sifs_
			- txtime(ETHER_ACK_LEN, basicRate_);
		*/
		timeout = txtime(ETHER_WAKE_ACK_LEN, basicRate_) 
			+ DSSS_LOW_MaxPropagationDelay;
		
		break;
	default:
		fprintf(stderr, "check_pktWAKE_ACK:Invalid MAC Control type\n");
		exit(1);
	}
	TRANSMIT(pktWAKE_ACK_, timeout);
	return 0;
}

// Checked
int
MacLow::check_pktWAKE()
{
	struct hdr_mac_low *mh;
	double timeout;

	assert(mhBackoff_.busy() == 0);

	if(pktWAKE_ == 0)
 		return -1;
	mh = HDR_MAC_LOW(pktWAKE_);

 	switch(mh->dh_fc.fc_type) {
	case MAC_LOW_Type_Wake:
		if(! is_idle()) {
			inc_cw();
			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		SET_TX_STATE(MAC_WAKE);
		if(ETHER_ADDR(mh->dh_da) != MAC_BROADCAST) {
		timeout = txtime(ETHER_WAKE_LEN, basicRate_)
			+ DSSS_LOW_MaxPropagationDelay			// XXX
			// XXX: New turnaround time here?
			// + sifs_
			+ DSSS_RxTxTurnaroundTime
			+ txtime(ETHER_WAKE_ACK_LEN, basicRate_)
			+ DSSS_LOW_MaxPropagationDelay;			// XXX
		}
		else {
			timeout = txtime(pktWAKE_) 
				+ DSSS_LOW_MaxPropagationDelay;
		}
		break;
	default:
		fprintf(stderr, "check_pktWAKE:Invalid MAC Control type\n");
		exit(1);
	}
	TRANSMIT(pktWAKE_, timeout);
	return 0;
}

/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
// Checked
void
MacLow::sendWAKE(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct wake_frame *wf = (struct wake_frame*)p->access(hdr_mac::offset_);
	
	assert(pktTx_);
	assert(pktWAKE_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_WAKE_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(wf, MAC_HDR_LEN);

	// wf->wf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	wf->wf_fc.fc_type	= MAC_LOW_Type_Wake;
	/*
 	wf->wf_fc.fc_subtype	= MAC_Subtype_RTS;
 	wf->wf_fc.fc_to_ds	= 0;
 	wf->wf_fc.fc_from_ds	= 0;
 	wf->wf_fc.fc_more_frag	= 0;
 	wf->wf_fc.fc_retry	= 0;
 	wf->wf_fc.fc_pwr_mgt	= 0;
 	wf->wf_fc.fc_more_data	= 0;
 	wf->wf_fc.fc_wep	= 0;
 	wf->wf_fc.fc_order	= 0;
	*/

	STORE4BYTE(&dst, (wf->wf_ra));
	
	/* store rts tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	STORE4BYTE(&index_, (wf->wf_ta));
	/* calculate rts duration field */
	/*
	wf->wf_duration = usec(sifs_
			       + txtime(ETHER_CTS_LEN, basicRate_)
			       + sifs_
			       + txtime(pktTx_)
			       + sifs_
			       + txtime(ETHER_ACK_LEN, basicRate_));
	*/
	wf->wf_duration = usec(
			       txtime(ETHER_WAKE_ACK_LEN, basicRate_)
			       );
	
	pktWAKE_ = p;
}

// Checked
void
MacLow::sendWAKE_ACK(int dst, double wake_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct wake_ack_frame *wa = (struct wake_ack_frame*)p->access(hdr_mac::offset_);

	assert(pktWAKE_ACK_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_WAKE_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;
	//ch->direction() = hdr_cmn::DOWN;
	bzero(wa, MAC_HDR_LEN);

	// cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	wa->wa_fc.fc_type	= MAC_LOW_Type_WakeAck;
	/*
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;
	*/
	
	//cf->cf_duration = CTS_DURATION(rts_duration);
	STORE4BYTE(&dst, (wa->wa_ra));
	
	/* store cts tx time */
	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate cts duration */
	/*
	wa->wa_duration = usec(sec(rts_duration)
			       - sifs_
			       - txtime(ETHER_CTS_LEN, basicRate_));
	*/
	// XXX; might need NAV for low and primary radio
	wa->wa_duration = 0;
	
	pktWAKE_ACK_ = p;
	
}

/* ======================================================================
   Retransmission Routines
   ====================================================================== */
// Checked
void
MacLow::RetransmitWAKE()
{
	assert(pktTx_);
	assert(pktWAKE_);
	assert(mhBackoff_.busy() == 0);
	struct wake_frame *wf = (struct wake_frame*)
			  pktWAKE_->access(hdr_mac::offset_);

	if(ETHER_ADDR(wf->wf_ra) == MAC_BROADCAST) {
		callbackPrimary(LOW_WAKE_TX);
		Packet::free(pktWAKE_); pktWAKE_ = 0;
		pktTx_ = 0;
		rst_cw();
		return;
	}

	macmib_->WAKEFailureCount++;

	ssrc_++;			// STA Short Retry Count

	if(ssrc_ >= macmib_->WakeRetryLimit) {
		discard(pktWAKE_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktWAKE_ = 0;
		callbackPrimary(LOW_DROP_PKT);
		/* tell the callback the send operation failed, let
		 * it handle dropping the packet */
		 pktTx_ = 0;
		ssrc_ = 0;
		rst_cw();
	} else {
		//printf("(%d)...retxing RTS:%x\n",index_,pktRTS_);
		struct wake_frame *wf;
		wf = (struct wake_frame*)pktWAKE_->access(hdr_mac::offset_);
		// wf->rf_fc.fc_retry = 1;

		inc_cw();
		mhBackoff_.start(cw_, is_idle());
	}
}

/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
// Checked
void
MacLow::send(Packet *p, Handler *h)
{
	struct hdr_mac_low* dh = HDR_MAC_LOW(p);

	/* 
	 * drop the packet if the node is in sleep mode
	 */
	if(netif_->is_sleeping() || netif_->is_off()) {
		return;
	}	

	assert(pktTx_ == 0);
	pktTx_ = p;
	
	callback_ = h;
	sendWAKE(ETHER_ADDR(dh->dh_da));

	if(mhBackoff_.busy() == 0) {
		if(is_idle()) {
			// If the media is idle, go ahead and send
			check_pktWAKE();
		}
		// If the media is not idle, start the backoff timer
		else {
			mhBackoff_.start(cw_, is_idle());
		}
	}

}

// Checked
void
MacLow::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	if(hdr->direction() == hdr_cmn::DOWN) {
                send(p, h);
                return;
	}
	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);
		pktRx_ = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(txtime(p));
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
		} else {
			collision(p);
		}
	}
}

// Checked
void
MacLow::recv_timer()
{
	u_int32_t src; 
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac_low *mh = HDR_MAC_LOW(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	// XXX debug
	//struct cts_frame *cf = (struct cts_frame*)pktRx_->access(hdr_mac::offset_);
	//u_int32_t src = ETHER_ADDR(mh->dh_sa);
	
	u_int8_t  type = mh->dh_fc.fc_type;

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);
	
	/*
	 *  If the interface is in TRANSMIT mode when this packet
	 *  "arrives", then I would never have seen it and should
	 *  do a silent discard without adjusting the NAV.
	 */
	if(tx_active_) {
	         Packet::free(pktRx_);
	        goto done;
	}

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == MAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);
		// XXX: May need to change this
		// set_nav(usec(eifs_));
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx_);
		// XXX: May need to change this
		// set_nav(usec(eifs_));
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
		set_nav(mh->dh_duration);
	}

	/*
	 * Address Filtering
	 */
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {
		case MAC_LOW_Type_Wake:
			recvWAKE(pktRx_);
			break;
		case MAC_LOW_Type_WakeAck:
			recvWAKE_ACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Type %x\n",
				type);
			exit(1);
	}
 done:
	pktRx_ = 0;
	rx_resume();
}

// Checked
void
MacLow::recvWAKE(Packet *p)
{
	assert(p != NULL);

	struct wake_frame *wf = (struct wake_frame*)p->access(hdr_mac::offset_);

	if(tx_state_ != MAC_IDLE) {
		fprintf(stderr, "tx_state not MAC_IDLE\n");
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this 
	 */
	if(pktWAKE_ACK_) {
		fprintf(stderr, "pktWAKE_ACK_ not null\n");
		discard(p, DROP_MAC_BUSY);
		return;
	}

	if(ETHER_ADDR(wf->wf_ra) != MAC_BROADCAST) {
#ifdef DEBUG
		fprintf(stderr, "%f %d in unicast recvWAKE\n", NOW, index_);
#endif
		sendWAKE_ACK(ETHER_ADDR(wf->wf_ta), wf->wf_duration);
		// Give a little delay
		if(!is_idle()) {
			mhDefer_.start(DSSS_LOW_RxTxTurnaroundTime);
			mac_log(p);
			return;
		}
	}
	else
		callbackPrimary(LOW_WAKE_RX);

	tx_resume();

	mac_log(p);
}

/*
 * txtime()	- pluck the precomputed tx time from the packet header
 */
double
MacLow::txtime(Packet *p)
 {
	 struct hdr_cmn *ch = HDR_CMN(p);
	 double t = ch->txtime();
	 if (t < 0.0) {
		 drop(p, "XXX");
 		exit(1);
	 }
	 return t;
 }

 
/*
 * txtime()	- calculate tx time for packet of size "psz" bytes 
 *		  at rate "drt" bps
 */
double
MacLow::txtime(double psz, double drt)
{
	double dsz = psz - PLCP_LOW_HDR_LEN;
	int plcp_hdr = PLCP_LOW_HDR_LEN << 3;
	int datalen = (int)dsz << 3;
	
	// double t = (((double)plcp_hdr)/phymib_->PLCPDataRate) + (((double)datalen)/drt);
	// For now, assume PLCP is transmitted at same speed as
	// data
	double t = (((double)plcp_hdr)/drt) + (((double)datalen)/drt);
	return(t);
}

// Checked
void
MacLow::recvWAKE_ACK(Packet *p)
{
	if(tx_state_ != MAC_WAKE) {
		fprintf(stderr, "%f %d dropping %s\n", NOW, index_, __FUNCTION__);
		return;
	}

	assert(pktWAKE_);
	Packet::free(pktWAKE_); pktWAKE_ = 0;
	pktTx_ = 0;

	callbackPrimary(LOW_WAKE_TX);

	assert(pktTx_);
	// debug
	//struct hdr_mac_low *mh = HDR_MAC_LOW(pktTx_);
	//printf("(%d):recvCTS:pktTx_-%x,mac-subtype-%d & pktCTS_:%x\n",index_,pktTx_,mh->dh_fc.fc_subtype,p);
	
	mhSend_.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */
	ssrc_ = 0;

	/*
	 * Backoff before sending again.
	 */
	rst_cw();
	assert(mhBackoff_.busy() == 0);
	mhBackoff_.start(cw_, is_idle());

	tx_resume();

	mac_log(p);
	// uptarget_->recv(p, (Handler*) 0);
}





