/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /home/mobiserve2/mjmille2/CVS_REPOSITORY/ns-2.26/mac/Attic/mac-low.h,v 1.1.2.9 2003/04/30 22:51:50 mjmille2 Exp $
 *
 */

#ifndef ns_mac_low_h
#define ns_mac_low_h

#include "mac-low-timers.h"
#include "marshall.h"
#include "mac-802_11m.h"
#include "mac-low_types.h"

// Defined in mac-802_11.h
/*
#define GET_ETHER_TYPE(x)		GET2BYTE((x))
#define SET_ETHER_TYPE(x,y)            {u_int16_t t = (y); STORE2BYTE(x,&t);}
*/

/* ======================================================================
   Frame Formats
   ====================================================================== */

class Mac802_11MHandler;
class Host;

struct wake_frame {
	struct frame_control_low	wf_fc;
	u_int16_t		wf_duration;
	u_char			wf_ra[ETHER_ADDR_LEN];
	u_char			wf_ta[ETHER_ADDR_LEN];
	u_char			wf_fcs[ETHER_FCS_LEN];
};

struct wake_ack_frame {
	struct frame_control_low	wa_fc;
	u_int16_t		wa_duration;
	u_char			wa_ra[ETHER_ADDR_LEN];
	u_char			wa_fcs[ETHER_FCS_LEN];
};


/* ======================================================================
   Definitions
   ====================================================================== */

#define PLCP_LOW_HDR_LEN                            \
  	((phymib_->PreambleLength >> 3) +        \
 	 (phymib_->PLCPHeaderLength >> 3))

/*
#define ETHER_HDR_LEN11				\
	(PLCP_HDR_LEN +				\
	 sizeof(struct hdr_mac_low) +		\
	 ETHER_FCS_LEN)
*/

#define ETHER_WAKE_LEN				\
        (PLCP_LOW_HDR_LEN +				\
	 sizeof(struct wake_frame))

#define ETHER_WAKE_ACK_LEN				\
         (PLCP_LOW_HDR_LEN +			\
	 sizeof(struct wake_ack_frame))

/*
 * IEEE 802.11 Spec, section 15.3.2
 *	- default values for the DSSS PHY MIB
 */
#define DSSS_LOW_CWMin			31
#define DSSS_LOW_CWMax			1023
#define DSSS_LOW_SlotTime			0.000020	// 20us
#define	DSSS_LOW_CCATime			0.000015	// 15us
#define DSSS_LOW_RxTxTurnaroundTime		0.000005	// 5us
#define DSSS_LOW_SIFSTime			0.000010	// 10us
// TODO: Change these values
// #define DSSS_LOW_PreambleLength		144		// 144 bits
#define DSSS_LOW_PreambleLength		32		// 144 bits
// #define DSSS_LOW_PLCPHeaderLength		48		// 48 bits
#define DSSS_LOW_PLCPHeaderLength		0		// 48 bits
// PLCPDataRate is not used
#define DSSS_LOW_PLCPDataRate		1.0e6		// 1Mbps

/* Must account for propagation delays added by the channel model when
 * calculating tx timeouts (as set in tcl/lan/ns-mac.tcl).
 *   -- Gavin Holland, March 2002
 */
#define DSSS_LOW_MaxPropagationDelay        0.000002        // 2us   XXXX

class PHY_LOW_MIB {
public:
	u_int32_t	CWMin;
	u_int32_t	CWMax;
	double		SlotTime;
	double		CCATime;
	double		RxTxTurnaroundTime;
	double		SIFSTime;
	u_int32_t	PreambleLength;
	u_int32_t	PLCPHeaderLength;
	double		PLCPDataRate;
};


/*
 * IEEE 802.11 Spec, section 11.4.4.2
 *      - default values for the MAC Attributes
 */
//#define MAC_LOW_RTSThreshold		3000		// bytes
#define MAC_LOW_RTSThreshold		0		// bytes
#define MAC_LOW_WakeRetryLimit		7		// retransmittions
#define MAC_LOW_LongRetryLimit		4		// retransmissions
#define MAC_LOW_FragmentationThreshold	2346		// bytes
#define MAC_LOW_MaxTransmitMSDULifetime	512		// time units
#define MAC_LOW_MaxReceiveLifetime		512		// time units

class MAC_LOW_MIB {
public:
	//		MACAddress;
	//		GroupAddresses;
	u_int32_t	RTSThreshold;
	u_int32_t	WakeRetryLimit;
	u_int32_t	LongRetryLimit;
	u_int32_t	FragmentationThreshold;
	u_int32_t	MaxTransmitMSDULifetime;
	u_int32_t	MaxReceiveLifetime;
	//		ManufacturerID;
	//		ProductID;

	u_int32_t	TransmittedFragmentCount;
	u_int32_t	MulticastTransmittedFrameCount;
	u_int32_t	FailedCount;	
	u_int32_t	RetryCount;
	u_int32_t	MultipleRetryCount;
	u_int32_t	FrameDuplicateCount;
	u_int32_t	RTSSuccessCount;
	u_int32_t	WAKEFailureCount;
	u_int32_t	ACKFailureCount;
	u_int32_t	ReceivedFragmentCount;
	u_int32_t	MulticastReceivedFrameCount;
	u_int32_t	FCSErrorCount;
};


/* ======================================================================
   The following destination class is used for duplicate detection.
   ====================================================================== */
// Defined in mac-802_11.h
/*
class Host {
public:
	LIST_ENTRY(Host) link;
	u_int32_t	index;
	u_int32_t	seqno;
};
*/

/* ======================================================================
   The actual 802.11 MAC class.
   ====================================================================== */
class MacLow : public Mac {
	friend class DeferTimerLow;
	friend class BackoffTimerLow;
	friend class IFTimerLow;
	friend class NavTimerLow;
	friend class RxTimerLow;
	friend class TxTimerLow;
	friend class Mac802_11M;
public:
	MacLow(PHY_LOW_MIB* p, MAC_LOW_MIB *m);
	void		recv(Packet *p, Handler *h);
	inline int	hdr_dst(char* hdr, int dst = -2);
	inline int	hdr_src(char* hdr, int src = -2);
	inline int	hdr_type(char* hdr, u_int16_t type = 0);
	double getWakeDelay(void);

protected:
	void	backoffHandler(void);
	void	deferHandler(void);
	void	navHandler(void);
	void	recvHandler(void);
	void	sendHandler(void);
	void	txHandler(void);

private:
	int		command(int argc, const char*const* argv);


	/*
	 * Called by the timers.
	 */
	void		recv_timer(void);
	void		send_timer(void);
	int		check_pktWAKE_ACK();
	int		check_pktWAKE();

	/*
	 * Packet Transmission Functions.
	 */
	void	send(Packet *p, Handler *h);
	void 	sendWAKE(int dst);
	void	sendWAKE_ACK(int dst, double wake_duration);
	void	RetransmitWAKE();

	/*
	 * Packet Reception Functions.
	 */
	void	recvWAKE(Packet *p);
	void	recvWAKE_ACK(Packet *p);

	void		capture(Packet *p);
	void		collision(Packet *p);
	void		discard(Packet *p, const char* why);
	void		rx_resume(void);
	void		tx_resume(void);
	void 		callbackPrimary(LowHandlerType lt);

	inline int	is_idle(void);

	/*
	 * Debugging Functions.
	 */
	void		trace_pkt(Packet *p);
	void		dump(char* fname);

	inline int initialized() {
		return (phymib_ && macmib_ && cache_ && logtarget_ && 
			Mac::initialized());
	}
	void mac_log(Packet *p) {
		logtarget_->recv(p, (Handler*) 0);
	}
	double txtime(Packet *p);
	double txtime(double psz, double drt);
	double txtime(int bytes) { /* clobber inherited txtime() */ abort(); }

	inline void inc_cw() {
		cw_ = (cw_ << 1) + 1;
		if(cw_ > phymib_->CWMax)
			cw_ = phymib_->CWMax;
	}
	inline void rst_cw() { cw_ = phymib_->CWMin; }
	inline double sec(double t) { return(t *= 1.0e-6); }
	inline u_int16_t usec(double t) {
		u_int16_t us = (u_int16_t)floor((t *= 1e6) + 0.5);
		/* u_int16_t us = (u_int16_t)rint(t *= 1e6); */
		return us;
	}
	inline void set_nav(u_int16_t us) {
		double now = Scheduler::instance().clock();
		double t = us * 1e-6;

		if((now + t) > nav_) {
			nav_ = now + t;
			if(mhNav_.busy())
				mhNav_.stop();
			mhNav_.start(t);
		}
	}

protected:
	PHY_LOW_MIB		*phymib_;
	MAC_LOW_MIB		*macmib_;

private:
	double		basicRate_;
 	double		dataRate_;
	
	/*
	 * Mac Timers
	 */
	IFTimerLow mhIF_;		// interface timer
	NavTimerLow	mhNav_;		// NAV timer
	RxTimerLow mhRecv_;		// incoming packets
	TxTimerLow mhSend_;		// outgoing packets
	DeferTimerLow mhDefer_;	// defer timer
	BackoffTimerLow mhBackoff_;	// backoff timer
	
	Mac802_11M *primaryMAC_;	

	// Optimization options
	bool rtsOpt;
	bool nextWakeOpt;

	/* ============================================================
	   Internal MAC State
	   ============================================================ */
	double		nav_;		// Network Allocation Vector

	MacState	rx_state_;	// incoming state (MAC_RECV or MAC_IDLE)
	MacState	tx_state_;	// outgoint state
	int		tx_active_;	// transmitter is ACTIVE

	Packet		*pktWAKE_;	// outgoing WAKE
	Packet		*pktWAKE_ACK_;	// outgoing WAKE-ACK

	u_int32_t	cw_;		// Contention Window
	u_int32_t	ssrc_;		// STA Short Retry Count
	/*
	double		sifs_;		// Short Interface Space
	double		pifs_;		// PCF Interframe Space
	double		difs_;		// DCF Interframe Space
	double		eifs_;		// Extended Interframe Space
	double		tx_sifs_;
	double		tx_pifs_;
	double		tx_difs_;
	*/

	int		min_frame_len_;

	NsObject*	logtarget_;

	/* ============================================================
	   Duplicate Detection state
	   ============================================================ */
	u_int16_t	sta_seqno_;	// next seqno that I'll use
	int		cache_node_count_;
	Host		*cache_;
};

#endif /* __mac_80211_h__ */

