#!/usr/bin/perl
use POSIX;

# num_nodes, pause time, max speed, sim time, max x, max y, 
# cbr/tcp, seed, num connections, rate

# These are not command line parameters
$pkt_size = 512;
$max_stop = 100;
$tcp_window = 32;

# Maximum communication range in meters
$max_range = 250;

# These are command line parameters
# These are the default values
$num_nodes = 10;
$pause = 100;
$max_speed = 20;
$stop_time  =100;
$max_x = 500;
$max_y = 500;
$type = "cbr";
$seed = 15934;
$num_conns = 5;
$rate = 2;
$cbrgen = 0; # Do the connections cbrgen style

# Helper function
sub max {
	if($_[0] > $_[1]) {
		return $_[0];
	}
	else {
		return $_[1];
	}
}

sub min {
	if($_[0] < $_[1]) {
		return $_[0];
	}
	else {
		return $_[1];
	}
}

# Get command line options
$args = join(" ", @ARGV);

$cbrgen_default = "FALSE";
if($cbrgen != 0) {
	$cbrgen_default = "TRUE";
}

$fname = "nn$num_nodes-x$max_x-y$max_y-ms$max_speed-p$pause-$type".
	"-c$num_conns-r$rate-rs$seed.scen";
$usage = "\nusage: 
	-f file name <DEFAULT: $fname>
	-nn number of nodes <DEFAULT: $num_nodes>
	-p random waypoint pause time <DEFAULT: $pause>
	-ms max movement speed (0 for no movement) <DEFAULT: $max_speed>
	-stop length of simulation (s) <DEFAULT: $stop_time> 
	-x max x coordinate <DEFAULT: $max_x> 
	-y max y coordinate <DEFAULT: $max_y>
	-t traffic type (cbr or tcp) <DEFAULT: $type>
	-rs random seed <DEFAULT: $seed>
	-c number of connections <DEFAULT: $num_conns>
	-r sending rate (pkts/sec) <DEFAULT: $rate>
	-cg do the connections like cbrgen <DEFAULT: $cbrgen_default>\n\n";

if(($tmp) = ($args =~ /.*-h.*/)) {
	die "$usage";
}

if(($tmp) = ($args =~ /.*-nn\s+(\d+).*/)) {
	$num_nodes = $tmp;
	if($num_nodes < 2) {
		die "nn must be greater than 1\n";
	}
}

if(($tmp) = ($args =~ /.*-p\s+(\d+\.*\d*).*/)) {
	$pause = $tmp;
}

if(($tmp) = ($args =~ /.*-ms\s+(\d+\.*\d*).*/)) {
	$max_speed = $tmp;
}

if(($tmp) = ($args =~ /.*-stop\s+(\d+\.*\d*).*/)) {
	$stop_time = $tmp;
	$max_stop = min($max_stop, $stop_time);
}

if(($tmp) = ($args =~ /.*-x\s+(\d+\.*\d*).*/)) {
	$max_x = $tmp;
}

if(($tmp) = ($args =~ /.*-y\s+(\d+\.*\d*).*/)) {
	$max_y = $tmp;
}

if(($tmp) = ($args =~ /.*-t\s+(cbr|tcp).*/)) {
	$type = $tmp;
}

if(($tmp) = ($args =~ /.*-rs\s+(\d+).*/)) {
	$seed = $tmp;
}

if(($tmp) = ($args =~ /.*-c\s+(\d+).*/)) {
	$num_conns = $tmp;
}

if(($tmp) = ($args =~ /.*-r\s+(\d+).*/)) {
	$rate = $tmp; # pkts per second
}

if(($tmp) = ($args =~ /.*-cg(?:|\s+.*)/)) {
	$cbrgen = 1;
}

$fname = "nn$num_nodes-x$max_x-y$max_y-ms$max_speed-p$pause-$type".
	"-c$num_conns-r$rate-rs$seed.scen";
if(($tmp) = ($args =~ /.*-f\s+(\w+).*/)) {
	$fname = $tmp;
}

if(0) {
print "nn $num_nodes\n";
print "p $pause\n";
print "ms $max_speed\n";
print "stop $stop_time\n";
print "x $max_x\n";
print "y $max_y\n";
print "t $type\n";
print "rs $seed\n";
print "c $num_conns\n";
print "r $rate\n";
print "cg $cbrgen\n";
}

open(out_file, ">$fname") || die "cannot open $fname\n";

print out_file "# nodes: $num_nodes, sim time: $stop_time, max x: $max_x\n";
print out_file "# max y: $max_y, max speed: $max_speed, pause: $pause\n";
print out_file "# traffic: $type, connections: $num_conns\n";
print out_file "# rate: $rate pkts/sec, seed: $seed\n\n";

@old_x; # prev x location
@old_y; # prev y location

@dist; # routing distances
@next_hop; # routing next hops

# infinity - must be greater than longest possible hop count
$infty = ($num_nodes + 1); 

$interval = (1/$rate);
$max_pkts = ($stop_time * $rate);

srand($seed);

# Set initial positions of nodes
sub position_nodes {
	for(my $i = 0; $i < $num_nodes; $i++) {
	
		my $x = rand $max_x;
		my $y = rand $max_y;
		my $z = 0;

		$old_x[$i] = $x;
		$old_y[$i] = $y;
	
		print out_file "\$node_($i) set X_ $x\n";
		print out_file "\$node_($i) set Y_ $y\n";
		print out_file "\$node_($i) set Z_ $z\n";
	
	}
}

sub create_cbr {

	my $src = $_[0];
	my $dst = $_[1];
	my $cbr_cnt = $_[2];
	my $stop = rand $max_stop;

	print out_file "#\n# $src connecting to $dst at time $stop\n#\n";
	print out_file "set udp_($cbr_cnt) \[new Agent/UDP\]\n";
	print out_file "\$ns_ attach-agent \$node_($src) \$udp_($cbr_cnt)\n";
	print out_file "set null_($cbr_cnt) \[new Agent/Null\]\n";
	print out_file "\$ns_ attach-agent \$node_($dst) \$null_($cbr_cnt)\n";
	print out_file "set cbr_($cbr_cnt) \[new Application/Traffic/CBR\]\n";
	print out_file "\$cbr_($cbr_cnt) attach-agent \$udp_($cbr_cnt)\n"; 
	print out_file "\$ns_ connect \$udp_($cbr_cnt) \$null_($cbr_cnt)\n";
#	print out_file "set cbr_($cbr_cnt) \[\$ns_ create-connection CBR ".
#		"\$node_($src) CBR \$node_($dst) 0\]\n";
	print out_file "\$cbr_($cbr_cnt) set packetSize_ $pkt_size\n";
	print out_file "\$cbr_($cbr_cnt) set interval_ $interval\n"; 
	print out_file "\$cbr_($cbr_cnt) set random_ 1\n"; 
	print out_file "\$cbr_($cbr_cnt) set maxpkts_ $max_pkts\n"; 

	print out_file "\$ns_ at $stop \"\$cbr_($cbr_cnt) start\"\n";

}

sub create_tcp {

	my $src = $_[0];
	my $dst = $_[1];
	my $tcp_cnt = $_[2];
	my $stop = rand $max_stop;

	print out_file "#\n# $src connecting to $dst at time $stop\n#\n";
	print out_file "set tcp_($tcp_cnt) \[\$ns_ create-connection ".
		"TCP \$node_($src) TCPSink \$node_($dst) 0\]\n";
	print out_file "\$tcp_($tcp_cnt) set window_ $tcp_window\n";
	print out_file "\$tcp_($tcp_cnt) set packetSize_ $pkt_size\n";
	print out_file "set ftp_($tcp_cnt) \[\$tcp_($tcp_cnt) attach-source FTP\]\n";
	
	print out_file "\$ns_ at $stop \"\$ftp_($tcp_cnt) start\"\n";

}

# Set connections
sub create_connections {
	my $conn_cnt = 0;
	
	if($cbrgen != 0) {
	for(my $i = 0; $i < $num_nodes; $i++) {
	
		my $tmp = int(rand 100);

		if($tmp < 50) { next; }

		my $src = $i;	
		my $dst = ($i + 1) % $num_nodes;
	
		if($type eq "cbr") {
			create_cbr($src, $dst, $conn_cnt);
		}
		else {
			create_tcp($src, $dst, $conn_cnt);
		}

		$conn_cnt++;
		if($conn_cnt == $num_conns) { last; }

		if($tmp < 75) { next; }
		
		my $dst = ($i + 2) % $num_nodes;
	
		if($type eq "cbr") {
			create_cbr($src, $dst, $conn_cnt);
		}
		else {
			create_tcp($src, $dst, $conn_cnt);
		}

		$conn_cnt++;
		if($conn_cnt == $num_conns) { last; }

	}
	}
	else {
	for(my $i = 0; $i < $num_conns; $i++) {
	
		my $src = int(rand $num_nodes);	
		my $dst = int(rand $num_nodes);
		while($src == $dst) {
			$dst = int(rand $num_nodes);
		}
	
		if($type eq "cbr") {
			create_cbr($src, $dst, $conn_cnt);
		}
		else {
			create_tcp($src, $dst, $conn_cnt);
		}
		$conn_cnt++;
	#	while(my $src_keys{$src} != undef) {
	#		$src = rand $num_nodes;
	#	}
	#	$src_keys{$src} = 1;	
		
	}
	}
}

sub in_range {

	my $i = $_[0];
	my $j = $_[1];
	my $euc_dist;

	my $x_diff = $old_x[$i] - $old_x[$j];
	my $y_diff = $old_y[$i] - $old_y[$j];

	$euc_dist = sqrt(($x_diff * $x_diff) + ($y_diff * $y_diff));

	if(0) {
	print "i: $i, j: $j\n";
	print "old_x: $old_x[$i] $old_x[$j]\n";
	print "old_y: $old_y[$i] $old_y[$j]\n";
	print "x_diff: $x_diff, y_diff: $y_diff\n";
	print "dist: $euc_dist, range: $max_range\n";
	}

	if($euc_dist < $max_range) {
		return 1;
	}
	return 0;
}

sub compute_neighbors {

	my $i, $j;

	for($i = 0; $i < $num_nodes; $i++) {
		for($j = $i; $j < $num_nodes; $j++) {
			if($i == $j) {
				$dist[$i][$j] = 0;
				$next_hop[$i][$j] = $i;
			}
			elsif(in_range($i,$j) == 1) {
				$dist[$i][$j] = 1;
				$dist[$j][$i] = 1;
				$next_hop[$i][$j] = $j;
				$next_hop[$j][$i] = $i;
			}
			else {
				$dist[$i][$j] = $infty;
				$dist[$j][$i] = $infty;
				$next_hop[$i][$j] = $infty;
				$next_hop[$j][$i] = $infty;
			}
		}
	}

}

# Get the shortest path routing tables for the
# topology.  Use Floyd-Warshall.
sub get_dists {

	my @prev_dist;
	my $i, $j, $k;

	compute_neighbors;

	for($k = 0; $k < $num_nodes; $k++) {

		for($i = 0; $i < $num_nodes; $i++) {
			$prev_dist[$i][$i] = $dist[$i][$i];
			for($j = ($i+1); $j < $num_nodes; $j++) {
				$prev_dist[$i][$j] = $dist[$j][$i];
				$prev_dist[$j][$i] = $dist[$i][$j];
			}
		}

		for($i = 0; $i < $num_nodes; $i++) {
			for($j = 0; $j < $num_nodes; $j++) {
				$old_dist = $prev_dist[$i][$j];
				$new_dist = ($prev_dist[$i][$k] + $prev_dist[$k][$j]);
				if($old_dist > $new_dist) {
					$dist[$i][$j] = $new_dist;
					$next_hop[$i][$j] = $next_hop[$i][$k];
				}
			}
		}
	}

}

sub print_dists {

	my $i, $j;

	for($i = 0; $i < $num_nodes; $i++) {
		for($j = $i; $j < $num_nodes; $j++) {
			print "$i -> $j: $dist[$i][$j] hops, $next_hop[$i][$j] next hop\n";
			print "$j -> $i: $dist[$j][$i] hops, $next_hop[$j][$i] next hop\n";
		}
	}

}

# Move the nodes
sub move_nodes {
	my @update;

	if($max_speed == 0) {
		return;
	}

	# Nodes randomly begin moving within pause time
	for(my $i = 0; $i < $num_nodes; $i++) {
		if($pause > 0) {
			$update[$i] = rand $pause;
		}
		else {
			$update[$i] = 0;
		}
	}

	my $finished = 0;

	while($finished == 0) {
		$finished = 1;
		for(my $i = 0; $i < $num_nodes; $i++) {
			if($update[$i] < $stop_time) {

				my $new_x = rand $max_x;
				my $new_y = rand $max_y;
				my $speed = rand $max_speed;

				print out_file "\$ns_ at $update[$i] \"\$node_($i) setdest ".
					"$new_x $new_y $speed\"\n";

				my $x_diff = ($new_x - $old_x[$i]);
				my $y_diff = ($new_y - $old_y[$i]);
				my $dist = sqrt(($x_diff * $x_diff) + ($y_diff * $y_diff));
				$old_x[$i] = $new_x;
				$old_y[$i] = $new_y;
				$update[$i] += ($pause + ($dist/$speed));

				$finished = 0;
			}
		}
	}
}

print "\n";
print "Positioning nodes...\n";
position_nodes;
#get_dists;
#print_dists;
print "Creating connections...\n";
create_connections;
print out_file "\n";
print "Calculating node movement...\n";
move_nodes;
print out_file "\n";

print "\nFinished!  File is:\n";
print "$fname\n\n";

close out_file;



